@@ -52,6 +52,10 @@ type TLSOptions struct {
 	ValidateHostname        bool
 }
 
+var (
+	errConnectionClosed = errors.New("connection closed")
+)
+
 // ConnectionListener is a user of a connection (eg. a producer or
 // a consumer) that can register itself to get notified
 // when the connection is closed.
@@ -89,6 +93,7 @@ type connectionState int32
 const (
 	connectionInit = iota
 	connectionReady
+	connectionClosing
 	connectionClosed
 )
 
@@ -98,6 +103,8 @@ func (s connectionState) String() string {
 		return "Initializing"
 	case connectionReady:
 		return "Ready"
+	case connectionClosing:
+		return "Closing"
 	case connectionClosed:
 		return "Closed"
 	default:
@@ -121,6 +128,7 @@ type incomingCmd struct {
 type connection struct {
 	sync.Mutex
 	cond              *sync.Cond
+	started           int32
 	state             ua.Int32
 	connectionTimeout time.Duration
 	closeOnce         sync.Once
@@ -135,13 +143,12 @@ type connection struct {
 
 	lastDataReceivedLock sync.Mutex
 	lastDataReceivedTime time.Time
-	pingTicker           *time.Ticker
-	pingCheckTicker      *time.Ticker
 
 	log log.Logger
 
 	requestIDGenerator uint64
 
+	incomingRequestsWG sync.WaitGroup
 	incomingRequestsCh chan *request
 	incomingCmdCh      chan *incomingCmd
 	closeCh            chan interface{}
@@ -149,7 +156,9 @@ type connection struct {
 
 	pendingLock sync.Mutex
 	pendingReqs map[uint64]*request
-	listeners   map[uint64]ConnectionListener
+
+	listenersLock sync.RWMutex
+	listeners     map[uint64]ConnectionListener
 
 	consumerHandlersLock sync.RWMutex
 	consumerHandlers     map[uint64]ConsumerHandler
@@ -181,8 +190,6 @@ func newConnection(opts connectionOptions) *connection {
 		log:                  opts.logger.SubLogger(log.Fields{"remote_addr": opts.physicalAddr}),
 		pendingReqs:          make(map[uint64]*request),
 		lastDataReceivedTime: time.Now(),
-		pingTicker:           time.NewTicker(keepAliveInterval),
-		pingCheckTicker:      time.NewTicker(keepAliveInterval),
 		tlsOptions:           opts.tls,
 		auth:                 opts.auth,
 
@@ -207,6 +214,11 @@ func newConnection(opts connectionOptions) *connection {
 }
 
 func (c *connection) start() {
+	if !atomic.CompareAndSwapInt32(&c.started, 0, 1) {
+		c.log.Warnf("connection has already started")
+		return
+	}
+
 	// Each connection gets its own goroutine that will
 	go func() {
 		if c.connect() {
@@ -215,11 +227,11 @@ func (c *connection) start() {
 				c.run()
 			} else {
 				c.metrics.ConnectionsHandshakeErrors.Inc()
-				c.changeState(connectionClosed)
+				c.Close()
 			}
 		} else {
 			c.metrics.ConnectionsEstablishmentErrors.Inc()
-			c.changeState(connectionClosed)
+			c.Close()
 		}
 	}()
 }
@@ -319,7 +331,7 @@ func (c *connection) waitUntilReady() error {
 	defer c.Unlock()
 
 	for c.getState() != connectionReady {
-		c.log.Debugf("Wait until connection is ready. State: %s", c.getState().String())
+		c.log.Debugf("Wait until connection is ready state=%s", c.getState().String())
 		if c.getState() == connectionClosed {
 			return errors.New("connection error")
 		}
@@ -331,32 +343,46 @@ func (c *connection) waitUntilReady() error {
 }
 
 func (c *connection) failLeftRequestsWhenClose() {
+	// wait for outstanding incoming requests to complete before draining
+	// and closing the channel
+	c.incomingRequestsWG.Wait()
+
 	reqLen := len(c.incomingRequestsCh)
 	for i := 0; i < reqLen; i++ {
 		c.internalSendRequest(<-c.incomingRequestsCh)
 	}
+
 	close(c.incomingRequestsCh)
 }
 
 func (c *connection) run() {
-	// All reads come from the reader goroutine
-	go c.reader.readFromConnection()
-	go c.runPingCheck()
-
-	c.log.Debugf("Connection run start channel %+v, requestLength %d", c, len(c.incomingRequestsCh))
+	pingSendTicker := time.NewTicker(keepAliveInterval)
+	pingCheckTicker := time.NewTicker(keepAliveInterval)
 
 	defer func() {
+		// stop tickers
+		pingSendTicker.Stop()
+		pingCheckTicker.Stop()
+
 		// all the accesses to the pendingReqs should be happened in this run loop thread,
-		// including the final cleanup, to avoid the issue https://github.com/apache/pulsar-client-go/issues/239
+		// including the final cleanup, to avoid the issue
+		// https://github.com/apache/pulsar-client-go/issues/239
 		c.pendingLock.Lock()
 		for id, req := range c.pendingReqs {
-			req.callback(nil, errors.New("connection closed"))
+			req.callback(nil, errConnectionClosed)
 			delete(c.pendingReqs, id)
 		}
 		c.pendingLock.Unlock()
 		c.Close()
 	}()
 
+	// All reads come from the reader goroutine
+	go c.reader.readFromConnection()
+	go c.runPingCheck(pingCheckTicker)
+
+	c.log.Debugf("Connection run starting with request capacity=%d queued=%d",
+		cap(c.incomingRequestsCh), len(c.incomingRequestsCh))
+
 	go func() {
 		for {
 			select {
@@ -380,30 +406,29 @@ func (c *connection) run() {
 
 		case cmd := <-c.incomingCmdCh:
 			c.internalReceivedCommand(cmd.cmd, cmd.headersAndPayload)
-
 		case data := <-c.writeRequestsCh:
 			if data == nil {
 				return
 			}
 			c.internalWriteData(data)
 
-		case <-c.pingTicker.C:
+		case <-pingSendTicker.C:
 			c.sendPing()
 		}
 	}
 }
 
-func (c *connection) runPingCheck() {
+func (c *connection) runPingCheck(pingCheckTicker *time.Ticker) {
 	for {
 		select {
 		case <-c.closeCh:
 			return
-		case <-c.pingCheckTicker.C:
+		case <-pingCheckTicker.C:
 			if c.lastDataReceived().Add(2 * keepAliveInterval).Before(time.Now()) {
 				// We have not received a response to the previous Ping request, the
 				// connection to broker is stale
 				c.log.Warn("Detected stale connection to broker")
-				c.TriggerClose()
+				c.Close()
 				return
 			}
 		}
@@ -445,7 +470,7 @@ func (c *connection) internalWriteData(data Buffer) {
 	c.log.Debug("Write data: ", data.ReadableBytes())
 	if _, err := c.cnx.Write(data.ReadableSlice()); err != nil {
 		c.log.WithError(err).Warn("Failed to write on connection")
-		c.TriggerClose()
+		c.Close()
 	}
 }
 
@@ -510,6 +535,9 @@ func (c *connection) internalReceivedCommand(cmd *pb.BaseCommand, headersAndPayl
 	case pb.BaseCommand_ERROR:
 		c.handleResponseError(cmd.GetError())
 
+	case pb.BaseCommand_SEND_ERROR:
+		c.handleSendError(cmd.GetError())
+
 	case pb.BaseCommand_CLOSE_PRODUCER:
 		c.handleCloseProducer(cmd.GetCloseProducer())
 
@@ -522,8 +550,6 @@ func (c *connection) internalReceivedCommand(cmd *pb.BaseCommand, headersAndPayl
 	case pb.BaseCommand_SEND_RECEIPT:
 		c.handleSendReceipt(cmd.GetSendReceipt())
 
-	case pb.BaseCommand_SEND_ERROR:
-
 	case pb.BaseCommand_MESSAGE:
 		c.handleMessage(cmd.GetMessage(), headersAndPayload)
 
@@ -536,7 +562,7 @@ func (c *connection) internalReceivedCommand(cmd *pb.BaseCommand, headersAndPayl
 
 	default:
 		c.log.Errorf("Received invalid command type: %s", cmd.Type)
-		c.TriggerClose()
+		c.Close()
 	}
 }
 
@@ -546,74 +572,85 @@ func (c *connection) Write(data Buffer) {
 
 func (c *connection) SendRequest(requestID uint64, req *pb.BaseCommand,
 	callback func(command *pb.BaseCommand, err error)) {
-	if c.getState() == connectionClosed {
+	c.incomingRequestsWG.Add(1)
+	defer c.incomingRequestsWG.Done()
+
+	state := c.getState()
+	if state == connectionClosed || state == connectionClosing {
 		callback(req, ErrConnectionClosed)
+
 	} else {
-		c.incomingRequestsCh <- &request{
+		select {
+		case <-c.closeCh:
+			callback(req, ErrConnectionClosed)
+
+		case c.incomingRequestsCh <- &request{
 			id:       &requestID,
 			cmd:      req,
 			callback: callback,
+		}:
 		}
 	}
 }
 
 func (c *connection) SendRequestNoWait(req *pb.BaseCommand) error {
-	if c.getState() == connectionClosed {
+	c.incomingRequestsWG.Add(1)
+	defer c.incomingRequestsWG.Done()
+
+	state := c.getState()
+	if state == connectionClosed || state == connectionClosing {
 		return ErrConnectionClosed
 	}
 
-	c.incomingRequestsCh <- &request{
+	select {
+	case <-c.closeCh:
+		return ErrConnectionClosed
+
+	case c.incomingRequestsCh <- &request{
 		id:       nil,
 		cmd:      req,
 		callback: nil,
+	}:
+		return nil
 	}
-	return nil
 }
 
 func (c *connection) internalSendRequest(req *request) {
-	c.pendingLock.Lock()
-	if req.id != nil {
-		c.pendingReqs[*req.id] = req
-	}
-	c.pendingLock.Unlock()
-	if c.getState() == connectionClosed {
+	if c.closed() {
 		c.log.Warnf("internalSendRequest failed for connectionClosed")
 		if req.callback != nil {
 			req.callback(req.cmd, ErrConnectionClosed)
 		}
 	} else {
+		c.pendingLock.Lock()
+		if req.id != nil {
+			c.pendingReqs[*req.id] = req
+		}
+		c.pendingLock.Unlock()
 		c.writeCommand(req.cmd)
 	}
 }
 
 func (c *connection) handleResponse(requestID uint64, response *pb.BaseCommand) {
-	c.pendingLock.Lock()
-	request, ok := c.pendingReqs[requestID]
+	request, ok := c.deletePendingRequest(requestID)
 	if !ok {
 		c.log.Warnf("Received unexpected response for request %d of type %s", requestID, response.Type)
-		c.pendingLock.Unlock()
 		return
 	}
 
-	delete(c.pendingReqs, requestID)
-	c.pendingLock.Unlock()
 	request.callback(response, nil)
 }
 
 func (c *connection) handleResponseError(serverError *pb.CommandError) {
 	requestID := serverError.GetRequestId()
-	c.pendingLock.Lock()
-	request, ok := c.pendingReqs[requestID]
+
+	request, ok := c.deletePendingRequest(requestID)
 	if !ok {
 		c.log.Warnf("Received unexpected error response for request %d of type %s",
 			requestID, serverError.GetError())
-		c.pendingLock.Unlock()
 		return
 	}
 
-	delete(c.pendingReqs, requestID)
-	c.pendingLock.Unlock()
-
 	errMsg := fmt.Sprintf("server error: %s: %s", serverError.GetError(), serverError.GetMessage())
 	request.callback(nil, errors.New(errMsg))
 }
@@ -621,14 +658,16 @@ func (c *connection) handleResponseError(serverError *pb.CommandError) {
 func (c *connection) handleSendReceipt(response *pb.CommandSendReceipt) {
 	producerID := response.GetProducerId()
 
-	c.Lock()
+	c.listenersLock.RLock()
 	producer, ok := c.listeners[producerID]
-	c.Unlock()
+	c.listenersLock.RUnlock()
 
 	if ok {
 		producer.ReceivedSendReceipt(response)
 	} else {
-		c.log.WithField("producerID", producerID).Warn("Got unexpected send receipt for message: ", response.MessageId)
+		c.log.
+			WithField("producerID", producerID).
+			Warn("Got unexpected send receipt for messageID=%+v", response.MessageId)
 	}
 }
 
@@ -648,6 +687,16 @@ func (c *connection) handleMessage(response *pb.CommandMessage, payload Buffer)
 	}
 }
 
+func (c *connection) deletePendingRequest(requestID uint64) (*request, bool) {
+	c.pendingLock.Lock()
+	defer c.pendingLock.Unlock()
+	request, ok := c.pendingReqs[requestID]
+	if ok {
+		delete(c.pendingReqs, requestID)
+	}
+	return request, ok
+}
+
 func (c *connection) lastDataReceived() time.Time {
 	c.lastDataReceivedLock.Lock()
 	defer c.lastDataReceivedLock.Unlock()
@@ -682,7 +731,7 @@ func (c *connection) handleAuthChallenge(authChallenge *pb.CommandAuthChallenge)
 	authData, err := c.auth.GetData()
 	if err != nil {
 		c.log.WithError(err).Warn("Failed to load auth credentials")
-		c.TriggerClose()
+		c.Close()
 		return
 	}
 
@@ -698,13 +747,35 @@ func (c *connection) handleAuthChallenge(authChallenge *pb.CommandAuthChallenge)
 	c.writeCommand(baseCommand(pb.BaseCommand_AUTH_RESPONSE, cmdAuthResponse))
 }
 
+func (c *connection) handleSendError(cmdError *pb.CommandError) {
+	c.log.Warnf("Received send error from server: [%v] : [%s]", cmdError.GetError(), cmdError.GetMessage())
+
+	requestID := cmdError.GetRequestId()
+
+	switch *cmdError.Error {
+	case pb.ServerError_NotAllowedError:
+		request, ok := c.deletePendingRequest(requestID)
+		if !ok {
+			c.log.Warnf("Received unexpected error response for request %d of type %s",
+				requestID, cmdError.GetError())
+			return
+		}
+
+		errMsg := fmt.Sprintf("server error: %s: %s", cmdError.GetError(), cmdError.GetMessage())
+		request.callback(nil, errors.New(errMsg))
+	case pb.ServerError_TopicTerminatedError:
+		// TODO: no-op
+	default:
+		// By default, for transient error, let the reconnection logic
+		// to take place and re-establish the produce again
+		c.Close()
+	}
+}
+
 func (c *connection) handleCloseConsumer(closeConsumer *pb.CommandCloseConsumer) {
 	consumerID := closeConsumer.GetConsumerId()
 	c.log.Infof("Broker notification of Closed consumer: %d", consumerID)
 
-	c.Lock()
-	defer c.Unlock()
-
 	if consumer, ok := c.consumerHandler(consumerID); ok {
 		consumer.ConnectionClosed()
 		c.DeleteConsumeHandler(consumerID)
@@ -717,78 +788,89 @@ func (c *connection) handleCloseProducer(closeProducer *pb.CommandCloseProducer)
 	c.log.Infof("Broker notification of Closed producer: %d", closeProducer.GetProducerId())
 	producerID := closeProducer.GetProducerId()
 
-	c.Lock()
-	defer c.Unlock()
-	if producer, ok := c.listeners[producerID]; ok {
-		producer.ConnectionClosed()
+	c.listenersLock.Lock()
+	producer, ok := c.listeners[producerID]
+	if ok {
 		delete(c.listeners, producerID)
+	}
+	c.listenersLock.Unlock()
+
+	// did we find a producer?
+	if ok {
+		producer.ConnectionClosed()
 	} else {
 		c.log.WithField("producerID", producerID).Warn("Producer with ID not found while closing producer")
 	}
 }
 
 func (c *connection) RegisterListener(id uint64, listener ConnectionListener) {
-	c.Lock()
-	defer c.Unlock()
+	// do not add if connection is closed
+	if c.closed() {
+		c.log.Warnf("Connection closed unable register listener id=%+v", id)
+		return
+	}
+
+	c.listenersLock.Lock()
+	defer c.listenersLock.Unlock()
 
 	c.listeners[id] = listener
 }
 
 func (c *connection) UnregisterListener(id uint64) {
-	c.Lock()
-	defer c.Unlock()
+	c.listenersLock.Lock()
+	defer c.listenersLock.Unlock()
 
 	delete(c.listeners, id)
 }
 
-// Triggers the connection close by forcing the socket to close and
-// broadcasting the notification on the close channel
-func (c *connection) TriggerClose() {
+// Close closes the connection by
+// closing underlying socket connection and closeCh.
+// This also triggers callbacks to the ConnectionClosed listeners.
+func (c *connection) Close() {
 	c.closeOnce.Do(func() {
+		c.Lock()
 		cnx := c.cnx
+		// do not use changeState() since they share the same lock
+		c.setState(connectionClosed)
+		c.cond.Broadcast()
+		c.Unlock()
+
 		if cnx != nil {
-			cnx.Close()
+			_ = cnx.Close()
 		}
 
 		close(c.closeCh)
+
+		listeners := make(map[uint64]ConnectionListener)
+		c.listenersLock.Lock()
+		for id, listener := range c.listeners {
+			listeners[id] = listener
+			delete(c.listeners, id)
+		}
+		c.listenersLock.Unlock()
+
+		consumerHandlers := make(map[uint64]ConsumerHandler)
+		c.consumerHandlersLock.Lock()
+		for id, handler := range c.consumerHandlers {
+			consumerHandlers[id] = handler
+			delete(c.consumerHandlers, id)
+		}
+		c.consumerHandlersLock.Unlock()
+
+		// notify producers connection closed
+		for _, listener := range listeners {
+			listener.ConnectionClosed()
+		}
+
+		// notify consumers connection closed
+		for _, handler := range consumerHandlers {
+			handler.ConnectionClosed()
+		}
+
+		c.metrics.ConnectionsClosed.Inc()
 	})
 }
 
-func (c *connection) Close() {
-	c.Lock()
-	defer c.Unlock()
-
-	c.cond.Broadcast()
-
-	if c.getState() == connectionClosed {
-		return
-	}
-
-	c.log.Info("Connection closed")
-	// do not use changeState() since they share the same lock
-	c.setState(connectionClosed)
-	c.TriggerClose()
-	c.pingTicker.Stop()
-	c.pingCheckTicker.Stop()
-
-	for _, listener := range c.listeners {
-		listener.ConnectionClosed()
-	}
-
-	consumerHandlers := make(map[uint64]ConsumerHandler)
-	c.consumerHandlersLock.RLock()
-	for id, handler := range c.consumerHandlers {
-		consumerHandlers[id] = handler
-	}
-	c.consumerHandlersLock.RUnlock()
-
-	for _, handler := range consumerHandlers {
-		handler.ConnectionClosed()
-	}
-
-	c.metrics.ConnectionsClosed.Inc()
-}
-
 func (c *connection) changeState(state connectionState) {
 	c.Lock()
 	c.setState(state)
@@ -804,6 +886,10 @@ func (c *connection) setState(state connectionState) {
 	c.state.Store(int32(state))
 }
 
+func (c *connection) closed() bool {
+	return connectionClosed == c.getState()
+}
+
 func (c *connection) newRequestID() uint64 {
 	return atomic.AddUint64(&c.requestIDGenerator, 1)
 }
@@ -843,6 +929,12 @@ func (c *connection) getTLSConfig() (*tls.Config, error) {
 }
 
 func (c *connection) AddConsumeHandler(id uint64, handler ConsumerHandler) {
+	// do not add if connection is closed
+	if c.closed() {
+		c.log.Warnf("Closed connection unable add consumer with id=%+v", id)
+		return
+	}
+
 	c.consumerHandlersLock.Lock()
 	defer c.consumerHandlersLock.Unlock()
 	c.consumerHandlers[id] = handler
